# Regular expression to match an ORDER BY clause with 'ASC' or 'DESC'.
ORDER_BY_QUERY = "(ASC|DESC)$"

# Regular expression to match any alphanumeric character (one or more times).
ANY_CHAR = "[a-zA-Z0-9_]+"

# Regular expression to match logical operators 'AND' or 'OR'.
OPERATOR = "(AND|OR)$"
